/*
 * Copyright (C) 2018 Inteno Broadband Technology AB. All rights reserved.
 *
 * Authors: Hrvoje Varga    <hrvoje.varga@sartura.hr>
 *          Matija Amidzic  <matija.amidzic@sartura.hr>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

#include <stdlib.h>
#include <libubox/uloop.h>
#include <libubox/list.h>
#include <libubox/runqueue.h>
#include <stdio.h>
#include <libubus.h>
#include <libubox/blobmsg.h>
#include <regex.h>

#include "monitor.h"
#include "debug.h"

static struct ubus_context *ubus_ctx;

typedef struct monitor_internal_s {
    monitor_t *monitor;
    int nr_times_matched;
    struct uloop_timeout timeout;
} monitor_internal_t;

typedef struct ubus_file_exec_s {
    int code;
    char *standard;
    char *error;
} ubus_file_exec_t;

static void timer_cb(struct uloop_timeout *t);
static int monitor_internal_start(struct list_head *monitors);

void (*monitor_test_cb)(char *command);
void (*monitor_execute_cb)(char *command);

int monitor_start(char *ubus_socket_path, struct list_head *monitors)
{
    _log("start monitoring");
    int error = 0;
    if (uloop_init()) {
        error = -1;
        goto exit;
    }

    // connect to ubus
    ubus_ctx = ubus_connect(ubus_socket_path);
    if (ubus_ctx == NULL) {
        error = -1;
        _error("failed to connect to ubus socket");
        goto exit;
    }

    monitor_internal_start(monitors);

exit:
    return error;
}

void monitor_stop(void)
{
    _debug("stop monitoring");
    uloop_cancelled = true;
    ubus_free(ubus_ctx);
}

typedef void(ubus_res_cb)(struct ubus_request *req, int type, struct blob_attr *msg);

static void monitor_ubus_call_file_exec(char *command, void *data, ubus_res_cb *cb)
{
    // perform ubus lookup
    uint32_t id;
    int rc = ubus_lookup_id(ubus_ctx, "file", &id);
    if (rc != UBUS_STATUS_OK) {
        // something went wrong
        _error("failed to perform ubus lookup");
        return;
    }

    // prepare ubus call
    static struct blob_buf buf = {0};
    blob_buf_init(&buf, 0);
    blobmsg_add_string(&buf, "command", "sh");
    void *params = blobmsg_open_array(&buf, "params");
    blobmsg_add_string(&buf, NULL, "-c");
    blobmsg_add_string(&buf, NULL, command);
    blobmsg_close_array(&buf, params);

    // perform ubus invoke
    ubus_invoke(ubus_ctx, id, "exec", buf.head, cb, data, 500);
    blob_buf_free(&buf);
}

static void monitor_system_call(char *command)
{
    int status;
    status = system(command);
    if (WIFEXITED(status)) {
        _log("system call exited with status '%d'", WEXITSTATUS(status));
    } else if (WIFSIGNALED(status)) {
        _log("system call was terminated with signal '%d'", WTERMSIG(status));
    } else {
        _log("system call exited abnormally");
    }
}

static void remove_trailing_newline(char *string)
{
    size_t string_len = strlen(string);
    string[string_len - 1] = '\0';
}

static void monitor_ubus_data_handler(struct ubus_request *req, int type, struct blob_attr *msg)
{
    (void) type; /* unused */
    _debug("monitor_ubus_data_handler called");
    enum {
        FILE_EXEC_CODE,
        FILE_EXEC_STDOUT,
        FILE_EXEC_STDERR,
        __FILE_EXEC_MAX,
    };

    static const struct blobmsg_policy policy[__FILE_EXEC_MAX] = {
            [FILE_EXEC_CODE] = {"code", BLOBMSG_TYPE_INT32},
            [FILE_EXEC_STDOUT] = {"stdout", BLOBMSG_TYPE_STRING},
            [FILE_EXEC_STDERR] = {"stderr", BLOBMSG_TYPE_STRING},
    };

    struct blob_attr *tb[__FILE_EXEC_MAX];

    if (blobmsg_parse(policy, __FILE_EXEC_MAX, tb, blobmsg_data(msg), blobmsg_len(msg)) != 0) {
        _error("parsing of ubus response failed");
        goto exit;
    }

    ubus_file_exec_t **file_exec = (ubus_file_exec_t **) req->priv;
    *file_exec = calloc(1, sizeof(ubus_file_exec_t));
    (*file_exec)->code = blobmsg_get_u32(tb[FILE_EXEC_CODE]);
    (*file_exec)->standard = blobmsg_get_string(tb[FILE_EXEC_STDOUT]);
    if ((*file_exec)->standard != NULL) {
        remove_trailing_newline((*file_exec)->standard);
    }
    (*file_exec)->error = blobmsg_get_string(tb[FILE_EXEC_STDERR]);
    if ((*file_exec)->error != NULL) {
        remove_trailing_newline((*file_exec)->error);
    }

exit:
    return;
}

static bool monitor_check_is_test_output_matched(char *test_output, monitor_t *monitor)
{
    bool is_test_matched = false;
    if (test_output == NULL) {
        _log("test output for monitor '%s' is empty", monitor->app);
        goto exit;
    }
    if (monitor->match_pattern == NULL) {
        _log("match_pattern for monitor '%s' is empty", monitor->app);
        goto exit;
    }

    switch (monitor->pattern_type) {
        case PATTERN_MATCH:
            if (strcmp(test_output, monitor->match_pattern) == 0) {
                _log("test output matches the string_match pattern configured for monitor '%s'", monitor->app);
                is_test_matched = true;
            }
            break;
        case PATTERN_NOT_MATCH:
            if (strcmp(test_output, monitor->match_pattern) != 0) {
                _log("test output matches the string_not_match pattern configured for monitor '%s'", monitor->app);
                is_test_matched = true;
            }
            break;
        case PATTERN_REGEX:;
            regex_t regex;
            int rc;
            rc = regcomp(&regex, monitor->match_pattern, REG_EXTENDED);
            if (rc) {
                char msgbuf[100] = {0};
                regerror(rc, &regex, msgbuf, sizeof(msgbuf));
                _error("regex compile for monitor '%s' failed: %s", monitor->app, msgbuf);
                break;
            }
            rc = regexec(&regex, test_output, 0, NULL, 0);
            if (!rc) {
                _log("test output matches the regex_match pattern configured for monitor '%s'", monitor->app);
                is_test_matched = true;
            } else if (rc != REG_NOMATCH) {
                char msgbuf[100] = {0};
                regerror(rc, &regex, msgbuf, sizeof(msgbuf));
                _error("regex match failed for monitor '%s': %s", monitor->app, msgbuf);
            }
            regfree(&regex);
            break;
        case PATTERN_UNSUPPORTED:
        default:
            _error("Unreachable code, pattern_type");
            abort();
    }

exit:
    return is_test_matched;
}

static void monitor_execute(monitor_t *monitor)
{
    _log("running execute for monitor '%s'", monitor->app);
    monitor_system_call(monitor->execute);

    // run custom function - for testing purposes
    if (monitor_execute_cb != NULL) {
        monitor_execute_cb(monitor->execute);
    }
}

static char *monitor_get_test_output_stream(ubus_file_exec_t *file_exec, monitor_t *monitor)
{
    monitor_stream_e stream = monitor->stream;
    switch (stream) {
        case STREAM_STDOUT:
            return file_exec->standard;
        case STREAM_STDERR:
            return file_exec->error;
        case STREAM_UNSUPPORTED:
            _error("invalid stream value, this should not happen!");
            return NULL;
    }
    return NULL;
}

static void timer_cb(struct uloop_timeout *t)
{
    monitor_internal_t *internal = container_of(t, monitor_internal_t, timeout);
    _debug("monitor rule name: %s, interval: %d", internal->monitor->app, internal->monitor->interval);

    // execute command based on test field
    ubus_file_exec_t *file_exec = NULL;
    monitor_ubus_call_file_exec(internal->monitor->test, &file_exec, monitor_ubus_data_handler);
    if (file_exec == NULL) {
        goto rearm_timer;
    }
    char *output = monitor_get_test_output_stream(file_exec, internal->monitor);

    // process the output from the test command
    bool is_test_matched = monitor_check_is_test_output_matched(output, internal->monitor);
    free(file_exec);
    if (is_test_matched == true) {
        internal->nr_times_matched++;
    }

    // run custom function - for testing purposes
    if (monitor_test_cb != NULL) {
        monitor_test_cb(internal->monitor->test);
    }

    // test should be matched nr_tests times before the condition is met.
    if (internal->nr_times_matched >= internal->monitor->nr_tests) {
        monitor_execute(internal->monitor);
        internal->nr_times_matched = 0;
    }

rearm_timer:
    uloop_timeout_set(&internal->timeout, internal->monitor->interval);
    uloop_timeout_add(&internal->timeout);
}

static int monitor_internal_start(struct list_head *monitors)
{
    int count = 0;
    monitor_t *monitor = NULL;
    list_for_each_entry(monitor, monitors, head)
    {
        count++;
    }
    _debug("number of monitors: %d", count);
    if (count == 0) {
        _log("no need to start monitoring as no monitors configured");
        return 0;
    }

    monitor_internal_t *internal_monitors = calloc(count, sizeof(*internal_monitors));

    monitor = NULL;
    int i = 0;
    list_for_each_entry(monitor, monitors, head)
    {
        internal_monitors[i].timeout.cb = timer_cb;
        uloop_timeout_set(&internal_monitors[i].timeout, monitor->interval);
        uloop_timeout_add(&internal_monitors[i].timeout);
        internal_monitors[i].nr_times_matched = 0;
        internal_monitors[i].monitor = monitor;
        i++;
    }

    uloop_run();
    uloop_done();
    free(internal_monitors);

    return 0;
}
