/*
 * Copyright (C) 2018 Inteno Broadband Technology AB. All rights reserved.
 *
 * Authors: Hrvoje Varga    <hrvoje.varga@sartura.hr>
 *          Matija Amidzic  <matija.amidzic@sartura.hr>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

#include <stdlib.h>
#include <limits.h>
#include <uci.h>
#include <libubox/list.h>

#include "debug.h"
#include "uci_parser.h"
#include "monitor.h"

#define UCI_NAME "imonitor"
#define UCI_MONITOR_SECTION "monitor"
#define UCI_MONITOR_APP_OPTION "app"
#define UCI_MONITOR_TEST_OPTION "test"
#define UCI_MONITOR_STREAM_OPTION "stream"
#define UCI_MONITOR_EXECUTE_OPTION "execute"
#define UCI_MONITOR_INTERVAL_OPTION "interval"
#define UCI_MONITOR_NR_TESTS_OPTION "nr_tests"

#define UCI_MONITOR_TOTAL_MATCH_OPTIONS 3
#define UCI_MONITOR_STRING_MATCH_OPTION "string_match"
#define UCI_MONITOR_STRING_NOT_MATCH_OPTION "string_not_match"
#define UCI_MONITOR_REGEX_MATCH_OPTION "regex_match"

#define DEFAULT_INTERVAL 5
#define INTERVAL_CONVERSION_FACTOR 1000 // 1000ms in 1s

static void uci_parser_free_monitor(monitor_t **monitor)
{
    if (*monitor != NULL) {
        if ((*monitor)->app != NULL) {
            free((*monitor)->app);
            (*monitor)->app = NULL;
        }
        if ((*monitor)->test != NULL) {
            free((*monitor)->test);
            (*monitor)->test = NULL;
        }
        if ((*monitor)->match_pattern != NULL) {
            free((*monitor)->match_pattern);
            (*monitor)->match_pattern = NULL;
        }
        if ((*monitor)->execute != NULL) {
            free((*monitor)->execute);
            (*monitor)->execute = NULL;
        }

        free(*monitor);
        *monitor = NULL;
    }
}

static parse_status_e uci_parser_parse_monitor_section(struct uci_context *uci_ctx, struct uci_section *s, monitor_t **monitor)
{
    parse_status_e status = PARSE_SUCCESS;

    _debug("parsing monitor section: %s", s->e.name);
    *monitor = calloc(1, sizeof(monitor_t));
    if ((*monitor) == NULL) {
        _error_memory("failed to allocate memory for storing monitor structure");
        status = PARSE_FAILED_INTERNAL_ERROR;
        goto exit;
    }

    // parse app option
    const char *option = uci_lookup_option_string(uci_ctx, s, UCI_MONITOR_APP_OPTION);
    if (option == NULL) {
        _error("%s: missing application name", s->e.name);
        status = PARSE_FAILED_MISSING_OPTION;
        goto free;
    }
    (*monitor)->app = strdup(option);
    if ((*monitor)->app == NULL) {
        _error_memory("failed to allocate memory for storing option");
        status = PARSE_FAILED_INTERNAL_ERROR;
        goto free;
    }
    _debug("app: %s", (*monitor)->app);

    // parse test option
    option = uci_lookup_option_string(uci_ctx, s, UCI_MONITOR_TEST_OPTION);
    if (option == NULL) {
        _error("%s: missing test command", s->e.name);
        status = PARSE_FAILED_MISSING_OPTION;
        goto free;
    }
    (*monitor)->test = strdup(option);
    if ((*monitor)->test == NULL) {
        _error_memory("failed to allocate memory for storing option");
        status = PARSE_FAILED_INTERNAL_ERROR;
        goto free;
    }
    _debug("test: %s", (*monitor)->test);

    // parse stream option
    option = uci_lookup_option_string(uci_ctx, s, UCI_MONITOR_STREAM_OPTION);
    if (option == NULL) {
        _error("%s: missing stream option", s->e.name);
        status = PARSE_FAILED_MISSING_OPTION;
        goto free;
    }
    if (strcmp(option, "stdout") == 0) {
        (*monitor)->stream = STREAM_STDOUT;
    } else if (strcmp(option, "stderr") == 0) {
        (*monitor)->stream = STREAM_STDERR;
    } else {
        _error("%s: unsupported stream type: %s", s->e.name, option);
        status = PARSE_FAILED_INVALID_VALUE;
        goto free;
    }
    _debug("stream: %d", (*monitor)->stream);

    // parse match_pattern options
    const char *match_pattern_option = NULL;
    monitor_match_pattern_e match_pattern_type = PATTERN_UNSUPPORTED;
    for (int i = 0; i < UCI_MONITOR_TOTAL_MATCH_OPTIONS; i++) {
        const char *temp_lookup_option = NULL;
        monitor_match_pattern_e temp_type = PATTERN_UNSUPPORTED;
        switch (i) {
            case 0:
                temp_lookup_option = UCI_MONITOR_STRING_MATCH_OPTION;
                temp_type = PATTERN_MATCH;
                break;
            case 1:
                temp_lookup_option = UCI_MONITOR_STRING_NOT_MATCH_OPTION;
                temp_type = PATTERN_NOT_MATCH;
                break;
            case 2:
                temp_lookup_option = UCI_MONITOR_REGEX_MATCH_OPTION;
                temp_type = PATTERN_REGEX;
                break;
            default:
                _error("unreachable code, match_pattern_option");
                abort();
        }

        const char *temp_match_option = NULL;
        temp_match_option = uci_lookup_option_string(uci_ctx, s, temp_lookup_option);

        // check that only one option is set
        if (match_pattern_option != NULL && temp_match_option != NULL) {
            _error("%s: invalid combination of options: only one of string_match, string_not_match or regex_match options must be set", s->e.name);
            status = PARSE_FAILED_INVALID_COMBINATION;
            goto free;
        }

        if (temp_match_option != NULL) {
            match_pattern_option = temp_match_option;
            match_pattern_type = temp_type;
        }
    }
    // check for a missing option
    if (match_pattern_option == NULL) {
        _error("%s: one of string_match, string_not_match or regex_match options must be set", s->e.name);
        status = PARSE_FAILED_MISSING_OPTION;
        goto free;
    }
    (*monitor)->match_pattern = strdup(match_pattern_option);
    if ((*monitor)->match_pattern == NULL) {
        _error_memory("failed to allocate memory for storing option");
        status = PARSE_FAILED_INTERNAL_ERROR;
        goto free;
    }
    (*monitor)->pattern_type = match_pattern_type;
    _debug("match_pattern: %s, type: %d", (*monitor)->match_pattern, (*monitor)->pattern_type);

    // parse execute option
    option = uci_lookup_option_string(uci_ctx, s, UCI_MONITOR_EXECUTE_OPTION);
    if (option == NULL) {
        _error("%s: missing execute command", s->e.name);
        status = PARSE_FAILED_MISSING_OPTION;
        goto free;
    }
    (*monitor)->execute = strdup(option);
    if ((*monitor)->execute == NULL) {
        _error_memory("failed to allocate memory for storing option");
        status = PARSE_FAILED_INTERNAL_ERROR;
        goto free;
    }
    _debug("execute: %s", (*monitor)->execute);

    // parse interval option
    option = uci_lookup_option_string(uci_ctx, s, UCI_MONITOR_INTERVAL_OPTION);
    double interval = DEFAULT_INTERVAL;
    if (option != NULL) {
        interval = strtod(option, NULL);
        // convert to ms
        interval = interval * INTERVAL_CONVERSION_FACTOR;
        if ((int) interval <= 0) {
            _error("%s: invalid interval value: %s", s->e.name, option);
            status = PARSE_FAILED_INVALID_VALUE;
            goto free;
        }

        // we store this value in an integer, so make sure it fits
        if (interval > INT_MAX) {
            interval = INT_MAX;
        }
    }
    (*monitor)->interval = (int) interval;
    _debug("interval: %d", (*monitor)->interval);

    // parse number of tests option
    option = uci_lookup_option_string(uci_ctx, s, UCI_MONITOR_NR_TESTS_OPTION);
    (*monitor)->nr_tests = 1;
    if (option != NULL) {
        int nr_tests = atoi(option);
        if (nr_tests <= 0) {
            _error("%s: invalid nr_tests value: %s", s->e.name, option);
            status = PARSE_FAILED_INVALID_VALUE;
            goto free;
        }
        (*monitor)->nr_tests = nr_tests;
    }
    _debug("nr_tests: %d", (*monitor)->nr_tests);

    goto exit;

free:
    uci_parser_free_monitor(monitor);

exit:
    return status;
}

parse_status_e uci_parser_parse(char *uci_config_path, struct list_head *monitors)
{
    int status = PARSE_SUCCESS;

    struct uci_context *uci_ctx = uci_alloc_context();
    if (NULL == uci_ctx) {
        _error("failed to allocate uci context");
        status = PARSE_FAILED_INTERNAL_ERROR;
        goto exit;
    }

    FILE *uci_config_file = NULL;
    uci_config_file = fopen(uci_config_path, "r");
    if (uci_config_file == NULL) {
        _error("failed to open uci config file: %s", uci_config_path);
        status = PARSE_FAILED_MISSING_FILE;
        goto free;
    }

    if (UCI_OK != uci_import(uci_ctx, uci_config_file, UCI_NAME, NULL, true)) {
        char *uci_error = NULL;
        uci_get_errorstr(uci_ctx, &uci_error, "uci");
        _error("uci lookup failed: details: %s", uci_error);
        free(uci_error);
        status = PARSE_FAILED_UCI_ERROR;
        goto free;
    }

    struct uci_ptr ptr;
    uci_lookup_ptr(uci_ctx, &ptr, (char *) UCI_NAME, false);

    struct uci_element *e = NULL;
    uci_foreach_element(&ptr.p->sections, e)
    {
        struct uci_section *s = uci_to_section(e);
        if (strcmp(s->type, UCI_MONITOR_SECTION) != 0) {
            _debug("ignoring uci section %s", s->type);
            continue;
        }

        monitor_t *monitor = NULL;
        status = uci_parser_parse_monitor_section(uci_ctx, s, &monitor);
        if (monitor == NULL) {
            goto free;
        }
        list_add(&monitor->head, monitors);
    }

free:
    uci_free_context(uci_ctx);
    if (uci_config_file != NULL) {
        fclose(uci_config_file);
    }

exit:
    return status;
}

void uci_parser_free_resources(struct list_head *monitors)
{
    monitor_t *prev_monitor = NULL;
    monitor_t *next_monitor = NULL;
    list_for_each_entry_safe(prev_monitor, next_monitor, monitors, head)
    {
        uci_parser_free_monitor(&prev_monitor);
    }
}
