/*
 * Copyright (C) 2018 Inteno Broadband Technology AB. All rights reserved.
 *
 * Authors:	Hrvoje Varga	<hrvoje.varga@sartura.hr>
 * 			Matija Amidzic	<matija.amidzic@sartura.hr>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

#include <stdarg.h>
#include <stddef.h>
#include <setjmp.h>
#include <cmocka.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include "control.h"
#include "monitor.h"

/** test_control_valid_uci_file
 *
 * Check that control will not return error in case when the uci file is valid.
 */
static void *test_control_valid_uci_file_init(void *vargp)
{
    int error = control_init("/var/run/ubus.sock", "data/uci_file_valid_execute_script");
    assert_int_equal(error, 0);
    return NULL;
}

static void *test_control_valid_uci_file_close(void *vargp)
{
    sleep(2);
    control_close();
    return NULL;
}

static void test_control_valid_uci_file(void **state)
{
    // setup the monitor callbacks
    monitor_test_cb = NULL;
    monitor_execute_cb = NULL;

    // start a new thread which will call `control_init` from control.h
    pthread_t control_init_tid;
    pthread_create(&control_init_tid, NULL, test_control_valid_uci_file_init, NULL);

    // start a new thread which will call `control_close` from control.h
    pthread_t control_close_tid;
    pthread_create(&control_close_tid, NULL, test_control_valid_uci_file_close, NULL);

    // wait for `control_close` to finish
    pthread_join(control_close_tid, NULL);

    // wait for `control_init` to finish
    pthread_join(control_init_tid, NULL);
}

/** test_control_invalid_uci_file
 *
 * Check that control will return an error in case when the uci file is invalid.
 */
static void test_control_invalid_uci_file(void **state)
{
    int error = control_init("/var/run/ubus.sock", "data/uci_file_invalid_format");
    assert_int_not_equal(error, 0);
}

/** test_control_execute_command_running
 *
 * Check that process is running after the execute is called.
 * Process is killed before and after the test to ensure the correctness of the test case.
 */
static int test_control_execute_command_running_setup(void **state)
{
    system("pkill cat");
    return 0;
}

static void *test_control_execute_command_running_init(void *vargp)
{
    int error = control_init("/var/run/ubus.sock", "data/uci_file_valid_execute_script");
    assert_int_equal(error, 0);
    return NULL;
}

static void *test_control_execute_command_running_close(void *vargp)
{
    sleep(2);
    control_close();
    return NULL;
}

static void test_control_execute_command_running(void **state)
{
    // setup the monitor callbacks
    monitor_test_cb = NULL;
    monitor_execute_cb = NULL;

    // start a new thread which will call `control_init` from control.h
    pthread_t control_init_tid;
    pthread_create(&control_init_tid, NULL, test_control_execute_command_running_init, NULL);

    // start a new thread which will call `control_close` from control.h
    pthread_t control_close_tid;
    pthread_create(&control_close_tid, NULL, test_control_execute_command_running_close, NULL);

    // wait for `control_close` to finish
    pthread_join(control_close_tid, NULL);

    // wait for `control_init` to finish
    pthread_join(control_init_tid, NULL);

    // check for running process
    char output[10] = {0};
    FILE *f = NULL;
    f = popen("pgrep cat", "r");
    if (f == NULL) {
        fail_msg("unable to run command");
        return;
    }
    fgets(output, sizeof(output) - 1, f);
    pclose(f);

    assert_string_not_equal(output, "");
}

static int test_control_execute_command_running_teardown(void **state)
{
    system("pkill cat");
    return 0;
}

int main(void)
{
    const struct CMUnitTest tests[] = {
        cmocka_unit_test(test_control_valid_uci_file),
        cmocka_unit_test(test_control_invalid_uci_file),
        cmocka_unit_test_setup_teardown(
            test_control_execute_command_running, test_control_execute_command_running_setup, test_control_execute_command_running_teardown),
    };
    return cmocka_run_group_tests(tests, NULL, NULL);
}
