/*
 * Copyright (C) 2018 Inteno Broadband Technology AB. All rights reserved.
 *
 * Authors:	Hrvoje Varga	<hrvoje.varga@sartura.hr>
 * 			Matija Amidzic	<matija.amidzic@sartura.hr>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

#include <stdarg.h>
#include <stddef.h>
#include <setjmp.h>
#include <cmocka.h>
#include <stdio.h>
#include <libubox/list.h>
#include "uci_parser.h"
#include "monitor.h"

#define TEST_DATA_PATH "data/"

/**
* test_uci_parser_missing_uci_file
*
* Check that uci_parser will return error PARSE_FAILED_MISSING_FILE in case when
* the path to non-existing uci configuration file is passed as an argument.
*/
static void test_uci_parser_missing_uci_file(void **state)
{
    (void) state; /* unused */
    struct list_head monitors = LIST_HEAD_INIT(monitors);
    parse_status_e status = uci_parser_parse("NON_EXISTING_UCI_FILE", &monitors);
    uci_parser_free_resources(&monitors);
    assert_int_equal(status, PARSE_FAILED_MISSING_FILE);
}

/**
* test_uci_parser_invalid_uci_format
*
* Check that uci-parser will return error PARSE_FAILED_UCI_ERROR in case when an
* invalid uci configuration file is passed as an argument.
* This invalid uci configuration file has an option without name and value.
*/
static void test_uci_parser_invalid_uci_format(void **state)
{
    (void) state; /* unused */
    struct list_head monitors = LIST_HEAD_INIT(monitors);
    parse_status_e status = uci_parser_parse(TEST_DATA_PATH "uci_file_invalid_format", &monitors);
    uci_parser_free_resources(&monitors);
    assert_int_equal(status, PARSE_FAILED_UCI_ERROR);
}

/**
* test_uci_parser_valid_uci_file
*
* Check that uci-parser will return PARSE_SUCCESS in case when a valid uci
* configuration file is passed as an argument.
* Test also checks wheather the file is correctly marshaled into an in-memory
* structure.
*/
static void test_uci_parser_valid_uci_file(void **state)
{
    (void) state; /* unused */
    struct list_head monitors = LIST_HEAD_INIT(monitors);
    parse_status_e status = uci_parser_parse(TEST_DATA_PATH "uci_file_valid", &monitors);
    assert_int_equal(status, PARSE_SUCCESS);
    assert_false(list_empty(&monitors));

    monitor_t *nodes[1];
    monitor_t *monitor = NULL;
    int size = 0;
    list_for_each_entry(monitor, &monitors, head)
    {
        nodes[size] = monitor;
        size++;
    }
    assert_int_equal(size, 1);

    assert_string_equal("bsd", nodes[0]->app);
    assert_string_equal("bsd -s", nodes[0]->test);
    assert_int_equal(STREAM_STDOUT, nodes[0]->stream);
    assert_string_equal("hangup", nodes[0]->match_pattern);
    assert_int_equal(PATTERN_MATCH, nodes[0]->pattern_type);
    assert_string_equal("killall -9 bsd; sleep 1; bsd", nodes[0]->execute);
    assert_int_equal(10000, nodes[0]->interval);
    assert_int_equal(2, nodes[0]->nr_tests);

    uci_parser_free_resources(&monitors);
}

/**
* test_uci_parser_float_interval_value
*
* Check that uci-parser will return PARSE_SUCCESS in case when a valid uci
* configuration file is passed as an argument.
* Test also checks wheather the file is correctly marshaled into an in-memory
* structure and that the floating point value for 'interval' option is accepted.
*/
static void test_uci_parser_float_interval_value(void **state)
{
    (void) state; /* unused */
    struct list_head monitors = LIST_HEAD_INIT(monitors);
    parse_status_e status = uci_parser_parse(TEST_DATA_PATH "uci_file_valid_float_interval", &monitors);
    assert_int_equal(status, PARSE_SUCCESS);
    assert_false(list_empty(&monitors));

    monitor_t *nodes[1];
    monitor_t *monitor = NULL;
    int size = 0;
    list_for_each_entry(monitor, &monitors, head)
    {
        nodes[size] = monitor;
        size++;
    }
    assert_int_equal(size, 1);

    assert_string_equal("bsd", nodes[0]->app);
    assert_string_equal("bsd -s", nodes[0]->test);
    assert_int_equal(STREAM_STDOUT, nodes[0]->stream);
    assert_string_equal("hangup", nodes[0]->match_pattern);
    assert_int_equal(PATTERN_MATCH, nodes[0]->pattern_type);
    assert_string_equal("killall -9 bsd; sleep 1; bsd", nodes[0]->execute);
    assert_int_equal(500, nodes[0]->interval);
    assert_int_equal(2, nodes[0]->nr_tests);

    uci_parser_free_resources(&monitors);
}
/**
* test_uci_parser_invalid_missing_app
*
* Check that uci-parser will return error PARSE_FAILED_MISSING_OPTION in case
* when an invalid uci configuration file is passed as an argument.
* This invalid uci configuration file is missing a mandatory `app` option.
*/
static void test_uci_parser_invalid_missing_app(void **state)
{
    (void) state; /* unused */
    struct list_head monitors = LIST_HEAD_INIT(monitors);
    parse_status_e status = uci_parser_parse(TEST_DATA_PATH "uci_file_invalid_missing_app", &monitors);
    uci_parser_free_resources(&monitors);
    assert_int_equal(status, PARSE_FAILED_MISSING_OPTION);
}

/**
* test_uci_parser_invalid_missing_test
*
* Check that uci-parser will return error PARSE_FAILED_MISSING_OPTION in case
* when an invalid uci configuration file is passed as an argument.
* This invalid uci configuration file is missing a mandatory `test` option.
*/
static void test_uci_parser_invalid_missing_test(void **state)
{
    (void) state; /* unused */
    struct list_head monitors = LIST_HEAD_INIT(monitors);
    parse_status_e status = uci_parser_parse(TEST_DATA_PATH "uci_file_invalid_missing_test", &monitors);
    uci_parser_free_resources(&monitors);
    assert_int_equal(status, PARSE_FAILED_MISSING_OPTION);
}

/**
* test_uci_parser_invalid_missing_execute
*
* Check that uci-parser will return error PARSE_FAILED_MISSING_OPTION in case
* when an invalid uci configuration file is passed as an argument.
* This invalid uci configuration file is missing a mandatory `execute` option.
*/
static void test_uci_parser_invalid_missing_execute(void **state)
{
    (void) state; /* unused */
    struct list_head monitors = LIST_HEAD_INIT(monitors);
    parse_status_e status = uci_parser_parse(TEST_DATA_PATH "uci_file_invalid_missing_execute", &monitors);
    uci_parser_free_resources(&monitors);
    assert_int_equal(status, PARSE_FAILED_MISSING_OPTION);
}

/**
* test_uci_parser_invalid_interval_value
*
* Check that uci-parser will return error PARSE_FAILED_INVALID_VALUE in case
* when an invalid uci configuration file is passed as an argument.
* This invalid uci configuration file is having an invalid value for `interval`
* option.
*/
static void test_uci_parser_invalid_interval_value(void **state)
{
    (void) state; /* unused */
    struct list_head monitors = LIST_HEAD_INIT(monitors);
    parse_status_e status = uci_parser_parse(TEST_DATA_PATH "uci_file_invalid_interval_value", &monitors);
    uci_parser_free_resources(&monitors);
    assert_int_equal(status, PARSE_FAILED_INVALID_VALUE);
}

/**
* test_uci_parser_invalid_negative_interval_value
*
* Check that uci-parser will return error PARSE_FAILED_INVALID_VALUE in case
* when an invalid uci configuration file is passed as an argument.
* This invalid uci configuration file is having a negative value for `interval`
* option.
*/
static void test_uci_parser_invalid_negative_interval_value(void **state)
{
    (void) state; /* unused */
    struct list_head monitors = LIST_HEAD_INIT(monitors);
    parse_status_e status = uci_parser_parse(TEST_DATA_PATH "uci_file_invalid_negative_interval_value", &monitors);
    uci_parser_free_resources(&monitors);
    assert_int_equal(status, PARSE_FAILED_INVALID_VALUE);
}

/**
* test_uci_parser_invalid_nr_tests_value
*
* Check that uci-parser will return error PARSE_FAILED_INVALID_VALUE in case
* when an invalid uci configuration file is passed as an argument.
* This invalid uci configuration file is having an invalid value for `nr_tests`
* option.
*/
static void test_uci_parser_invalid_nr_tests_value(void **state)
{
    (void) state; /* unused */
    struct list_head monitors = LIST_HEAD_INIT(monitors);
    parse_status_e status = uci_parser_parse(TEST_DATA_PATH "uci_file_invalid_nr_tests_value", &monitors);
    uci_parser_free_resources(&monitors);
    assert_int_equal(status, PARSE_FAILED_INVALID_VALUE);
}

/**
* test_uci_parser_invalid_missing_stream
*
* Check that uci-parser will return error PARSE_FAILED_MISSING_OPTION in case
* when an invalid uci configuration file is passed as an argument.
* This invalid uci configuration file is missing a mandatory `stream` option.
*/
static void test_uci_parser_invalid_missing_stream(void **state)
{
    (void) state; /* unused */
    struct list_head monitors = LIST_HEAD_INIT(monitors);
    parse_status_e status = uci_parser_parse(TEST_DATA_PATH "uci_file_invalid_missing_stream", &monitors);
    uci_parser_free_resources(&monitors);
    assert_int_equal(status, PARSE_FAILED_MISSING_OPTION);
}

/**
* test_uci_parser_invalid_missing_string_match
*
* Check that uci-parser will return error PARSE_FAILED_MISSING_OPTION in case
* when an invalid uci configuration file is passed as an argument.
* This invalid uci configuration file is missing a mandatory `string_match`
* option.
*/
static void test_uci_parser_invalid_missing_string_match(void **state)
{
    (void) state; /* unused */
    struct list_head monitors = LIST_HEAD_INIT(monitors);
    parse_status_e status = uci_parser_parse(TEST_DATA_PATH "uci_file_invalid_missing_string_match", &monitors);
    uci_parser_free_resources(&monitors);
    assert_int_equal(status, PARSE_FAILED_MISSING_OPTION);
}

/**
* test_uci_parser_invalid_stream_value
*
* Check that uci-parser will return error PARSE_FAILED_INVALID_VALUE in case
* when an invalid uci configuration file is passed as an argument.
* This invalid uci configuration file is having an invalid value for `stream`
* option.
*/
static void test_uci_parser_invalid_stream_value(void **state)
{
    (void) state; /* unused */
    struct list_head monitors = LIST_HEAD_INIT(monitors);
    parse_status_e status = uci_parser_parse(TEST_DATA_PATH "uci_file_invalid_stream_value", &monitors);
    uci_parser_free_resources(&monitors);
    assert_int_equal(status, PARSE_FAILED_INVALID_VALUE);
}

/**
* test_uci_parser_ignore_unsupported_section
*
* Check that uci-parser will ignore unsupported sections.
*/
static void test_uci_parser_ignore_unsupported_section(void **state)
{
    (void) state; /* unused */
    struct list_head monitors = LIST_HEAD_INIT(monitors);
    parse_status_e status = uci_parser_parse(TEST_DATA_PATH "uci_file_valid_unsupported_section", &monitors);
    int count = 0;
    monitor_t *monitor = NULL;
    list_for_each_entry(monitor, &monitors, head)
    {
        count++;
    }
    assert_int_equal(count, 1);
    uci_parser_free_resources(&monitors);
    assert_int_equal(status, PARSE_SUCCESS);
}

/**
* test_uci_parser_string_not_match
*
* Check that uci-parser supports a string_not_match option.
*/
static void test_uci_parser_string_not_match(void **state)
{
    (void) state; /* unused */
    struct list_head monitors = LIST_HEAD_INIT(monitors);
    parse_status_e status = uci_parser_parse(TEST_DATA_PATH "uci_file_valid_string_not_match", &monitors);
    assert_int_equal(status, PARSE_SUCCESS);
    assert_false(list_empty(&monitors));

    monitor_t *nodes[1];
    monitor_t *monitor = NULL;
    int size = 0;
    list_for_each_entry(monitor, &monitors, head)
    {
        nodes[size] = monitor;
        size++;
    }
    assert_int_equal(size, 1);

    assert_string_equal("not matched string", nodes[0]->match_pattern);
    assert_int_equal(PATTERN_NOT_MATCH, nodes[0]->pattern_type);

    uci_parser_free_resources(&monitors);
}

/**
* test_uci_parser_regex_match
*
* Check that uci-parser supports a regex_match option.
*/
static void test_uci_parser_regex_match(void **state)
{
    (void) state; /* unused */
    struct list_head monitors = LIST_HEAD_INIT(monitors);
    parse_status_e status = uci_parser_parse(TEST_DATA_PATH "uci_file_valid_regex_match", &monitors);
    assert_int_equal(status, PARSE_SUCCESS);
    assert_false(list_empty(&monitors));

    monitor_t *nodes[1];
    monitor_t *monitor = NULL;
    int size = 0;
    list_for_each_entry(monitor, &monitors, head)
    {
        nodes[size] = monitor;
        size++;
    }
    assert_int_equal(size, 1);

    assert_string_equal(".*test.*", nodes[0]->match_pattern);
    assert_int_equal(PATTERN_REGEX, nodes[0]->pattern_type);

    uci_parser_free_resources(&monitors);
}

/**
* test_uci_parser_both_string_match_and_string_not_match
*
* Check that uci-parser reports and error in case when both string_match and string_not_match are present.
*/
static void test_uci_parser_both_string_match_and_string_not_match(void **state)
{
    (void) state; /* unused */
    struct list_head monitors = LIST_HEAD_INIT(monitors);
    parse_status_e status = uci_parser_parse(TEST_DATA_PATH "uci_file_invalid_both_string_match_and_string_not_match", &monitors);
    assert_int_equal(status, PARSE_FAILED_INVALID_COMBINATION);
    assert_true(list_empty(&monitors));
}

/**
* test_uci_parser_both_string_match_and_regex_match
*
* Check that uci-parser reports and error in case when both string_match and regex_match are present.
*/
static void test_uci_parser_both_string_match_and_regex_match(void **state)
{
    (void) state; /* unused */
    struct list_head monitors = LIST_HEAD_INIT(monitors);
    parse_status_e status = uci_parser_parse(TEST_DATA_PATH "uci_file_invalid_both_string_match_and_regex_match", &monitors);
    assert_int_equal(status, PARSE_FAILED_INVALID_COMBINATION);
    assert_true(list_empty(&monitors));
}

/**
* test_uci_parser_string_match_string_not_match_regex_match
*
* Check that uci-parser reports and error in case when all match types are present.
*/
static void test_uci_parser_string_match_string_not_match_regex_match(void **state)
{
    (void) state; /* unused */
    struct list_head monitors = LIST_HEAD_INIT(monitors);
    parse_status_e status = uci_parser_parse(TEST_DATA_PATH "uci_file_invalid_string_match_string_not_match_regex_match", &monitors);
    assert_int_equal(status, PARSE_FAILED_INVALID_COMBINATION);
    assert_true(list_empty(&monitors));
}

int main(void)
{
    const struct CMUnitTest tests[] = {
        cmocka_unit_test(test_uci_parser_missing_uci_file),
        cmocka_unit_test(test_uci_parser_invalid_uci_format),
        cmocka_unit_test(test_uci_parser_valid_uci_file),
        cmocka_unit_test(test_uci_parser_float_interval_value),
        cmocka_unit_test(test_uci_parser_invalid_missing_app),
        cmocka_unit_test(test_uci_parser_invalid_missing_test),
        cmocka_unit_test(test_uci_parser_invalid_missing_execute),
        cmocka_unit_test(test_uci_parser_invalid_interval_value),
        cmocka_unit_test(test_uci_parser_invalid_negative_interval_value),
        cmocka_unit_test(test_uci_parser_invalid_nr_tests_value),
        cmocka_unit_test(test_uci_parser_invalid_missing_stream),
        cmocka_unit_test(test_uci_parser_invalid_missing_string_match),
        cmocka_unit_test(test_uci_parser_invalid_stream_value),
        cmocka_unit_test(test_uci_parser_ignore_unsupported_section),
        cmocka_unit_test(test_uci_parser_string_not_match),
        cmocka_unit_test(test_uci_parser_regex_match),
        cmocka_unit_test(test_uci_parser_both_string_match_and_string_not_match),
        cmocka_unit_test(test_uci_parser_both_string_match_and_regex_match),
        cmocka_unit_test(test_uci_parser_string_match_string_not_match_regex_match),

    };
    return cmocka_run_group_tests(tests, NULL, NULL);
}
